/* Straightforward POSIX server for Linux nbd (Network Block Device). */

/* Copyright 2004 Andrew Isaacson <adi@hexapodia.org> */
/* This program is Free Software licensed under the GNU GPL version 2.0 */
/* see file COPYING in this directory for more information */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

typedef unsigned char		u8;
typedef unsigned int		u32;
typedef unsigned long long	u64;

/* do not allow requests for more than 256MB */
#define MAX_BYTES	(1024*1024*256)

/* definitions taken from linux-2.4.25/include/linux/nbd.h */

#define NBD_REQ_MAGIC	0x25609513
#define NBD_REP_MAGIC	0x67446698

/* from linux-2.4.25/include/linux/fs.h */
#define READ 0
#define WRITE 1
#define READA 2		/* read-ahead  - don't block if no resources */
#define SPECIAL 4	/* For non-blockdevice requests in request queue */

/* all values in network byte order */
typedef struct nbd_request {
    u32 magic;		/* must be NBD_REQ_MAGIC */
    u32 cmd;		/* READ or WRITE */
    u64 handle;		/* opaque identifier */
    u64 off;		/* file byte position */
    u32 len;		/* size of request */
    			/* ... if WRITE, len bytes follow */
} nbd_request;

typedef struct nbd_reply {
    u32 magic;		/* must be NBD_REP_MAGIC */
    u32 err;		/* 0 = success, anything else = error */
    u64 handle;		/* copied from request */
    			/* ... if READ, len bytes follow */
} nbd_reply;

int o_rdonly = 0, o_debug = 0;

#define N 10

struct nbd_request reqs[N];
int next_req;

#define htonu64 ntohu64
#define ntohu32 ntohl
#define htonu32 htonl

u64 ntohu64(u64 v)
{
    static int need = -1;

    if(need == -1) {
	u64 a, b = 0xfee1b0da;
	a = ((u64)htonl(b & 0xffffffff) << 32) | htonl(b >> 32);
	need = a != b;
	assert(htonu64(ntohu64(0x11223344556677ULL)) == 0x11223344556677ULL);
    }
    if(!need) return v;
    return ((u64)htonl(v & 0xffffffff) << 32) | htonl(v >> 32);
}

void die(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}

void debug(int, char *, ...) __attribute__((format(printf, 2, 3)));

void debug(int severity, char *fmt, ...)
{
    va_list ap;

    if(severity > o_debug) return;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
}

void usage(const char *prog)
{
    die("usage: %s [-r] port file [size]\n", prog);
}

off_t fdsize(int fd)
{
    struct stat s;
    fstat(fd, &s);
    if(S_ISCHR(s.st_mode) || S_ISBLK(s.st_mode)) return -1;
    return s.st_size;
}

ssize_t write_all(int fd, const void *buf, size_t sz)
{
    const char *p = buf;
    int n, nzero = 0;

    do {
	if((n = write(fd, p, sz)) == -1)
	    return n;
	if(n == 0 && ++nzero > 5)
	    die("write(%d, %p, %ld) = 0\n", fd, buf, (long)sz);
	sz -= n;
	p += n;
    } while(sz > 0);
    return p - (const char *)buf;
}

ssize_t read_all(int fd, void *buf, size_t sz)
{
    char *p = buf;
    ssize_t n;

    do {
	if(((n = read(fd, p, sz)) == -1) || (n == 0))
	    return n;
	sz -= n;
	p += n;
    } while(sz > 0);
    return p - (const char *)buf;
}

int wait_connection(int port)
{
    struct sockaddr_in sa;
    socklen_t ssz;
    int sock, x = 1;

    if((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1)
	die("socket: %s\n", strerror(errno));
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);
    sa.sin_addr.s_addr = INADDR_ANY;
    if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x)) == -1)
	die("setsockopt: %s\n", strerror(errno));
    if(bind(sock, (struct sockaddr *)&sa, sizeof(sa)) == -1)
	die("bind: %s\n", strerror(errno));
    if(listen(sock, 2) == -1)
	die("listen: %s\n", strerror(errno));
    ssz = sizeof(sa);
    debug(1, "waiting for connection on port %d\n", port);
    if((sock = accept(sock, (struct sockaddr*)&sa, &ssz)) == -1)
	die("accept: %s\n", strerror(errno));
    debug(1, "connection from %s:%d on fd %d\n",
	inet_ntoa(sa.sin_addr), ntohs(sa.sin_port), sock);
    return sock;
}

u64 nbd_magic = 0x00420281861253LL;
#define INIT_PASSWD "NBDMAGIC"

/* derived from the requirements of nbd-client.c from nbd-2.7.1 */
/* XXX what's the 128 bytes for? */
int send_hello(int sock, off_t sz)
{
    char *p, buf[sizeof(INIT_PASSWD)-1 + 2*sizeof(u64) + 128] = { 0 };
    int n = sizeof(buf);
    u64 c;

    p = buf;
    memcpy(p, INIT_PASSWD, sizeof(INIT_PASSWD)-1);
    p += sizeof(INIT_PASSWD)-1;
    c = htonu64(nbd_magic);
    memcpy(p, &c, sizeof(c));
    p += sizeof(c);
    c = htonu64(sz);
    memcpy(p, &c, sizeof(c));

    debug(2, "sending %d bytes of hello\n", n);
    if(write_all(sock, buf, n) == -1)
	die("hello: write(): %s\n", strerror(errno));
    return 0;
}

void ntoh_request(nbd_request *r)
{
    r->magic  = ntohu32(r->magic);
    r->cmd    = ntohu32(r->cmd);
    r->off    = ntohu64(r->off);
    r->len    = ntohu32(r->len);
}

void hton_reply(nbd_reply *r)
{
    r->magic = htonu32(r->magic);
    r->err   = htonu32(r->err);
}

void hexdump(FILE *f, void *p, int n)
{
    u8 *a = p;
    int i;

    for(i=0; i<n; i++)
	fprintf(f, "%s%02x%s", i % 16 == 0 ? "    " : i % 8 == 0 ? " " : "",
	    a[i],
	    (i % 16 == 15 || i == n-1) ? "\n" : " ");
}

char *magicstr(u32 magic)
{
    if(magic == NBD_REQ_MAGIC) return "REQ_MAGIC";
    if(magic == NBD_REP_MAGIC) return "REP_MAGIC";
    return "BAD MAGIC";
}

char *cmdstr(u32 cmd)
{
    if(cmd == READ) return "Read";
    if(cmd == WRITE) return "Write";
    if(cmd == READA) return "ReadA";
    if(cmd == SPECIAL) return "Special";
    return "BAD CMD";
}

void dump_request(FILE *f, nbd_request *r)
{
    fprintf(f, "req@%p: %s %s %d@%lld (0x%x@0x%llx) 0x%llx\n",
	r, magicstr(r->magic), cmdstr(r->cmd),
	r->len, r->off, r->len, r->off, r->handle);
    hexdump(f, r, sizeof(*r));
}

void *get_req(int sock)
{
    struct nbd_request *r = &reqs[next_req];
    u8 *p = (void *)r;
    int n, sz = sizeof(*r);

    debug(3, "waiting to fill req slot %d\n", next_req);
    next_req = (next_req + 1) % N;
    do {
	if((n = read(sock, p, sz)) == -1)
	    die("read: %s\n", strerror(errno));
	if(n == 0) die("eof on socket\n");
	debug(3, "read %d bytes\n", n);
	sz -= n;
	p += n;
    } while(sz);
    if(o_debug > 3) hexdump(stderr, r, sizeof(*r));
    ntoh_request(r);
    if(r->magic != NBD_REQ_MAGIC) {
	dump_request(stderr, r);
	die("bad magic 0x%08x!\n", r->magic);
    }
    return r;
}

int err(nbd_request *r, int sock, int ev)
{
    nbd_reply rep;

    rep.magic = NBD_REP_MAGIC;
    rep.handle = r->handle;
    rep.err = ev;
    hton_reply(&rep);
    debug(1, "handle 0x%llx err %d (%s)\n", r->handle, ev, strerror(ev));
    write_all(sock, &rep, sizeof(rep));
    return -1;
}

int handle_write(nbd_request *r, nbd_reply *rep, int sock, int fd, off_t fdsz)
{
    char *buf;
    int rc;

    if(o_rdonly) return err(r, sock, EPERM);
    if(!(buf = malloc(r->len)))
	return err(r, sock, ENOMEM);
    read_all(sock, buf, r->len);
    rep->err = 0;
    debug(2, "write(%d, %p, %d (0x%x))", fd, buf, r->len, r->len);
    if((rc = write_all(fd, buf, r->len)) == -1)
	rep->err = errno;
    debug(2, " = %d %s\n", rc, rc == -1 ? strerror(errno) : "");
    hton_reply(rep);
    write_all(sock, rep, sizeof(*rep));
    return 0;
}

int handle_read(nbd_request *r, nbd_reply *rep, int sock, int fd, off_t fdsz)
{
    char *buf;
    size_t len;

    len = sizeof(*rep) + r->len;
    if(!(buf = malloc(len)))
	return err(r, sock, ENOMEM);
    rep->err = 0;
    if(r->len) {
	if(read(fd, buf + sizeof(*rep), r->len) == -1)
	    rep->err = errno;
    }
    hton_reply(rep);
    memcpy(buf, rep, sizeof(*rep));
    write_all(sock, buf, len);
    return 0;
}

int handle_reada(nbd_request *r, nbd_reply *rep, int sock, int fd, off_t fdsz)
{
    return err(r, sock, EINVAL);
}

int handle_special(nbd_request *r, nbd_reply *rep, int sock, int fd, off_t fdsz)
{
    return err(r, sock, EINVAL);
}

int handle_req(nbd_request *r, int sock, int fd, off_t fdsz)
{
    nbd_reply rep;

    rep.magic = NBD_REP_MAGIC;
    rep.handle = r->handle;
    if(r->len > MAX_BYTES)
	return err(r, sock, EINVAL);
    if(r->off > fdsz || r->off + r->len > fdsz)
	return err(r, sock, EFAULT);
    if(lseek(fd, r->off, SEEK_SET) == -1)
	return err(r, sock, errno);
    switch(r->cmd) {
    case WRITE:	return handle_write(r, &rep, sock, fd, fdsz);
    case READ:	return handle_read(r, &rep, sock, fd, fdsz);
    case READA:	return handle_reada(r, &rep, sock, fd, fdsz);
    case SPECIAL: return handle_special(r, &rep, sock, fd, fdsz);
    default:	return err(r, sock, EINVAL);
    }
}

void client(int port)
{
    char cmd[40];
    /* simple test program, connect to localhost on 'port' and send a few
       requests */
    FILE *p;
    u8 r[] = {
	0x25, 0x60, 0x95, 0x13,				/* magic */
	0x00, 0x00, 0x00, 0x01,				/* type */
	0x61, 0x64, 0x69, 0x40, 0x68, 0x65, 0x78, 0x61, /* handle */
	0x42, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* offset */
	0x0e, 0xee, 0xee, 0xee
    };

    /* -w1: wait just 1 second for timeouts; a hack, because netcat doesn't
     * close the socket when getting EOF on stdin
     */
    snprintf(cmd, sizeof(cmd), "nc -w1 localhost %d | hexdump", port);
    p = popen(cmd, "w");
    if(!p) die("unable to popen(nc)\n");
    fwrite(r, sizeof(r), 1, p);
    if(fclose(p) != 0)
	die("fclose: %s\n", strerror(errno));
    exit(0);
}

int main(int argc, char **argv)
{
    int c;
    int port, fd, sock;
    off_t sz;
    void *r;

    while((c = getopt(argc, argv, "drt:")) != -1) {
	switch(c) {
	case 'd': o_debug++; break;
	case 'r': o_rdonly = 1; break;
	case 't': port = atoi(optarg); client(port); break;
	default: usage(argv[0]);
	}
    }

    if(!(argv[optind] && argv[optind+1])) usage(argv[0]);
    port = atoi(argv[optind]);
    if((fd = open(argv[optind+1], o_rdonly ? O_RDONLY : O_RDWR)) == -1)
	die("%s: %s\n", argv[optind+1], strerror(errno));
    sz = fdsize(fd);
    if(sz == -1 && optind+2 >= argc)
	die("Must provide size when referencing device special file\n");
    if(argv[optind+2]) {
	off_t s = strtoll(argv[optind+2], 0, 0);
	if(sz != (off_t)-1 && s > sz)
	    die("Requested length %lld > actual length %lld\n", s, sz);
	sz = s;
    }

    sock = wait_connection(port);
    send_hello(sock, sz);
    while((r = get_req(sock)) != 0) {
	if(o_debug > 3) dump_request(stderr, r);
	handle_req(r, sock, fd, sz);
    }
    return 0;
}
